@@ -18,7 +18,6 @@ You should have received a copy of the GNU General Public License
 along with GAS; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
-#include <stdio.h>
 #include <ctype.h>
 
 #include "m68k-opcode.h"
@@ -45,7 +44,7 @@ long omagic = OMAGIC;
 
 /* This array holds the chars that always start a comment.  If the
    pre-processor is disabled, these aren't very useful */
-char comment_chars[] = "|";
+const char comment_chars[] = "|";
 
 /* This array holds the chars that only start a comment at the beginning of
    a line.  If the line seems to have the form '# 123 filename'
@@ -54,16 +53,16 @@ char comment_chars[] = "|";
    first line of the input file.  This is because the compiler outputs
    #NO_APP at the beginning of its output. */
 /* Also note that '/*' will always start a comment */
-char line_comment_chars[] = "#";
+const char line_comment_chars[] = "#";
 
 /* Chars that can be used to separate mant from exp in floating point nums */
-char EXP_CHARS[] = "eE";
+const char EXP_CHARS[] = "eE";
 
 /* Chars that mean this number is a floating point constant */
 /* As in 0f12.456 */
 /* or    0d1.2345e12 */
 
-char FLT_CHARS[] = "rRsSfFdDxXeEpP";
+const char FLT_CHARS[] = "rRsSfFdDxXeEpP";
 
 /* Also be aware that MAXIMUM_NUMBER_OF_CHARS_FOR_FLOAT may have to be
    changed in read.c .  Ideally it shouldn't have to know about it at all,
@@ -90,6 +89,7 @@ struct obstack robyn;
 #define PCREL		3
 #define BCC68000        4
 #define DBCC            5
+#define PCLEA		6
 
 /* BCC68000 is for patching in an extra jmp instruction for long offsets
    on the 68000.  The 68000 doesn't support long branches with branchs */
@@ -106,7 +106,7 @@ struct obstack robyn;
 	How many bytes this mode will add to the size of the frag
 	Which mode to go to if the offset won't fit in this one
  */
-relax_typeS
+const relax_typeS
 md_relax_table[] = {
 { 1,		1,		0,	0 },	/* First entries aren't used */
 { 1,		1,		0,	0 },	/* For no good reason except */
@@ -114,29 +114,34 @@ md_relax_table[] = {
 { 1,		1,		0,	0 },
 
 { (127),	(-128),		0,	TAB(BRANCH,SHORT)},
-{ (2+32767),	(2+-32768),	2,	TAB(BRANCH,LONG) },
+{ (32767),	(-32768),	2,	TAB(BRANCH,LONG) },
 { 0,		0,		4,	0 },
 { 1,		1,		0,	0 },
 
 { 1,		1,		0,	0 },	/* FBRANCH doesn't come BYTE */
-{ (2+32767),	(2+-32768),	2,	TAB(FBRANCH,LONG)},
+{ (32767),	(-32768),	2,	TAB(FBRANCH,LONG)},
 { 0,		0,		4,	0 },
 { 1,		1,		0,	0 },
 
 { 1,		1,		0,	0 },	/* PCREL doesn't come BYTE */
-{ (2+32767),	(2+-32768),	2,	TAB(PCREL,LONG)},
+{ (32767),	(-32768),	2,	TAB(PCREL,LONG)},
 { 0,		0,		4,	0 },
 { 1,		1,		0,	0 },
 
-{ (127),        (-128),         0,      TAB(BCC68000,SHORT)},
-{ (2+32767),    (2+-32768),     2,      TAB(BCC68000,LONG) },
-{ 0,            0,              6,      0 },  /* jmp long space */
-{ 1,            1,              0,      0 },
+{ (127),	(-128),		0,	TAB(BCC68000,SHORT)},
+{ (32767),	(-32768),	2,	TAB(BCC68000,LONG) },
+{ 0,		0,		6,	0 },	/* jmp long space */
+{ 1,		1,		0,	0 },
 
-{ 1,            1,              0,      0 },  /* DBCC doesn't come BYTE */
-{ (2+32767),    (2+-32768),     2,      TAB(DBCC,LONG) },
-{ 0,            0,             10,      0 },  /* bra/jmp long space */
-{ 1,            1,              0,      0 },
+{ 1,		1,		0,	0 },	/* DBCC doesn't come BYTE */
+{ (32767),	(-32768),	2,	TAB(DBCC,LONG) },
+{ 0,		0,		10,	0 },	/* bra/jmp long space */
+{ 1,		1,		0,	0 },
+
+{ 1,		1,		0,	0 },	/* PCLEA doesn't come BYTE */
+{ 32767,	-32768,		2,	TAB(PCLEA,LONG) },
+{ 0,		0,		6,	0 },
+{ 1,		1,		0,	0 },
 
 };
 
@@ -150,11 +155,11 @@ void	s_proc(),	float_cons();
 
 /* This table describes all the machine specific pseudo-ops the assembler
    has to support.  The fields are:
-   	  pseudo-op name without dot
+ 	  pseudo-op name without dot
 	  function to call to execute this pseudo-op
 	  Integer arg to pass to the function
  */
-pseudo_typeS md_pseudo_table[] = {
+const pseudo_typeS md_pseudo_table[] = {
 	{ "data1",	s_data1,	0	},
 	{ "data2",	s_data2,	0	},
 	{ "even",	s_even,		0	},
@@ -176,7 +181,7 @@ pseudo_typeS md_pseudo_table[] = {
 #define isword(x)	((x)>=-32768 && (x)<=65535)
 #define islong(x)	(1)
 
-char *input_line_pointer;
+extern char *input_line_pointer;
 
 /* Operands we can parse:  (And associated modes)
 
@@ -442,13 +447,21 @@ register char **ccp;
 		c3,
 		c4;
 	register int n = 0,
-		ret;
+		ret = FAIL;
 
 	c1=mklower(ccp[0][0]);
+#ifdef REGISTER_PREFIX
+	if(c1!=REGISTER_PREFIX)
+		return FAIL;
+	c1=mklower(ccp[0][1]);
+	c2=mklower(ccp[0][2]);
+	c3=mklower(ccp[0][3]);
+	c4=mklower(ccp[0][4]);
+#else
 	c2=mklower(ccp[0][1]);
 	c3=mklower(ccp[0][2]);
 	c4=mklower(ccp[0][3]);
-
+#endif
 	switch(c1) {
 	case 'a':
 		if(c2>='0' && c2<='7') {
@@ -522,7 +535,7 @@ register char **ccp;
 		break;
 	case 'f':
 		if(c2=='p') {
- 			if(c3>='0' && c3<='7') {
+			if(c3>='0' && c3<='7') {
 				n=3;
 				ret = FPREG+c3-'0';
 				if(c4==':')
@@ -628,6 +641,9 @@ register char **ccp;
 		break;
 	}
 	if(n) {
+#ifdef REGISTER_PREFIX
+		n++;
+#endif
 		if(isalnum(ccp[0][n]) || ccp[0][n]=='_')
 			ret=FAIL;
 		else
@@ -694,25 +710,6 @@ register struct m68k_op *opP;
 			return OK;
 		}
 		opP->mode=ABSL;
-		if(strend[-1]==':') {	/* mode ==foo:[wl] */
-			switch(*strend) {
-			case 'w':
-			case 'W':
-				opP->isiz=2;
-				break;
-			case 'l':
-			case 'L':
-				opP->isiz=3;
-				break;
-			default:
-				opP->error="size spec not :w or :l";
-				return FAIL;
-			}
-			strend-=2;
-		} else {
-			opP->isiz=0;
-		}
-
 		opP->con1=add_exp(str,strend);
 		return OK;
 	}
@@ -767,7 +764,30 @@ register struct m68k_op *opP;
 			case ')': --i; break;
 			}
 		}
-		opP->con1=add_exp(beg_str,str-2);
+		/* if(str[-3]==':') {
+			int siz;
+
+			switch(str[-2]) {
+			case 'b':
+			case 'B':
+				siz=1;
+				break;
+			case 'w':
+			case 'W':
+				siz=2;
+				break;
+			case 'l':
+			case 'L':
+				siz=3;
+				break;
+			default:
+				opP->error="Specified size isn't :w or :l";
+				return FAIL;
+			}
+			opP->con1=add_exp(beg_str,str-4);
+			opP->con1->e_siz=siz;
+		} else */
+			opP->con1=add_exp(beg_str,str-2);
 			/* Should be offset,reg */
 		if(str[-1]==',') {
 			i=try_index(&str,opP);
@@ -871,10 +891,17 @@ struct m68k_op *opP;
 		return FAIL;
 	}
 	SKIP_W();
-	if(mklower(*ss)=='w') opP->isiz=2;
-	else if(mklower(*ss)=='l') opP->isiz=3;
-	else {
-		opP->error="Size spec not :w or :l";
+	switch(*ss) {
+	case 'w':
+	case 'W':
+		opP->isiz=2;
+		break;
+	case 'l':
+	case 'L':
+		opP->isiz=3;
+		break;
+	default:
+		opP->error="Index register size spec not :w or :l";
 		*s=ss;
 		return FAIL;
 	}
@@ -938,7 +965,7 @@ static struct hash_control*   op_hash = NULL;	/* handle of the OPCODE hash table
 
 
 /*
- *                  m 6 8 _ i p ( )
+ *		m 6 8 _ i p ( )
  *
  * This converts a string into a 68k instruction.
  * The string must be a bare single instruction in sun format
@@ -980,7 +1007,6 @@ char	*instring;
 	int	losing;
 	int	opsfound;
 	char	*crack_operand();
-	char	*atof_m68k();
 	LITTLENUM_TYPE words[6];
 	LITTLENUM_TYPE *wordp;
 
@@ -1061,7 +1087,7 @@ char	*instring;
 
 			case '#':
 				if(opP->mode!=IMMED)
- 					losing++;
+					losing++;
 				else {
 					long t;
 
@@ -1433,34 +1459,52 @@ char	*instring;
 				tmpreg=0x18+opP->reg-ADDR; /* 3.areg */
 				break;
 			case AOFF:
-				if(opP->reg==PC)
-					tmpreg=0x3A; /* 7.2 */
-				else
-					tmpreg=0x28+opP->reg-ADDR; /* 5.areg */
+
 				nextword=get_num(opP->con1,80);
 				/* Force into index mode.  Hope this works */
-				if(!issword(nextword)) {
+
+				/* We do the first bit for 32-bit displacements,
+				   and the second bit for 16 bit ones.  It is
+				   possible that we should make the default be
+				   WORD instead of LONG, but I think that'd
+				   break GCC, so we put up with a little
+				   inefficiency for the sake of working output.
+				 */
+
+				if(   !issword(nextword)
+				   || (   isvar(opP->con1)
+				       && (  (   opP->con1->e_siz==0
+					      && flagseen['l']==0)
+					   || opP->con1->e_siz==3))) {
+
 					if(opP->reg==PC)
 						tmpreg=0x3B;	/* 7.3 */
 					else
 						tmpreg=0x30+opP->reg-ADDR;	/* 6.areg */
-					/* addword(0x0171); */
-/* 171 seems to be wrong, and I can't find the 68020 manual, so we'll try 170
-   (which is what the Sun asm seems to generate */
-					addword(0x0170);
-					if(isvar(opP->con1))
-						add_fix('l',opP->con1,0);
-					addword(nextword>>16);
-					/* addword(nextword); */
-				} else if(isvar(opP->con1)) {
-					if(!flagseen['l']) {
-						add_fix('w',opP->con1,0);
-					} else {
-						tmpreg=0x30+opP->reg-ADDR;
-						addword(0x0170);
-						add_fix('l',opP->con1,0);
-						addword(nextword>>16);
+					if(isvar(opP->con1)) {
+						if(opP->reg==PC) {
+						       add_frag(adds(opP->con1),
+							 offs(opP->con1),
+							 TAB(PCLEA,SZ_UNDEF));
+							break;
+						} else {
+							addword(0x0170);
+							add_fix('l',opP->con1,1);
+						}
 					}
+					addword(nextword>>16);
+				} else {
+					if(opP->reg==PC)
+						tmpreg=0x3A; /* 7.2 */
+					else
+						tmpreg=0x28+opP->reg-ADDR; /* 5.areg */
+
+					if(isvar(opP->con1)) {
+						if(opP->reg==PC) {
+							add_fix('w',opP->con1,1);
+						} else
+							add_fix('w',opP->con1,0);
+						}
 				}
 				addword(nextword);
 				break;
@@ -1498,12 +1542,20 @@ char	*instring;
 					case 8: nextword|=0x600; break;
 					default: abort();
 					}
-						/* IF its simple, GET US OUT OF HERE! */
-						/* Must be INDEX, with an index register.  Address register
-						   cannot be ZERO-PC, and either :b was forced, or we know it'll fit */
-					if(opP->mode==AINDX &&
- opP->reg!=FAIL && opP->reg!=ZPC && (siz1==1 || (issbyte(baseo) &&
- !isvar(opP->con1)))) {
+						/* IF its simple,
+						   GET US OUT OF HERE! */
+
+						/* Must be INDEX, with an index
+						   register.  Address register
+						   cannot be ZERO-PC, and either
+						   :b was forced, or we know
+						   it will fit */
+					if(   opP->mode==AINDX
+					   && opP->reg!=FAIL
+					   && opP->reg!=ZPC
+					   && (   siz1==1
+					       || (   issbyte(baseo)
+						   && !isvar(opP->con1)))) {
 						nextword +=baseo&0xff;
 						addword(nextword);
 						if(isvar(opP->con1))
@@ -1512,7 +1564,7 @@ char	*instring;
 					}
 				} else
 					nextword|=0x40;	/* No index reg */
-				
+
 					/* It aint simple */
 				nextword|=0x100;
 					/* If the guy specified a width, we assume that
@@ -1568,15 +1620,25 @@ char	*instring;
 				}
 				addword(nextword);
 
-				if(isvar(opP->con1))
-					add_fix(siz1==3 ? 'l' : 'w',opP->con1,0);
+				if(isvar(opP->con1)) {
+					if(opP->reg==PC || opP->reg==ZPC) {
+						add_fix(siz1==3 ? 'l' : 'w',opP->con1,1);
+						opP->con1->e_exp.X_add_number+=6;
+					} else
+						add_fix(siz1==3 ? 'l' : 'w',opP->con1,0);
+				}
 				if(siz1==3)
 					addword(baseo>>16);
 				if(siz1)
 					addword(baseo);
 
-				if(isvar(opP->con2))
-					add_fix(siz2==3 ? 'l' : 'w',opP->con2,0);
+				if(isvar(opP->con2)) {
+					if(opP->reg==PC || opP->reg==ZPC) {
+						add_fix(siz2==3 ? 'l' : 'w',opP->con2,1);
+						opP->con1->e_exp.X_add_number+=6;
+					} else
+						add_fix(siz2==3 ? 'l' : 'w',opP->con2,0);
+				}
 				if(siz2==3)
 					addword(outro>>16);
 				if(siz2)
@@ -1586,7 +1648,9 @@ char	*instring;
 
 			case ABSL:
 				nextword=get_num(opP->con1,80);
-				switch(opP->isiz) {
+				switch(opP->con1->e_siz) {
+				default:
+					as_warn("Unknown size for absolute reference");
 				case 0:
 					if(!isvar(opP->con1) && issword(offs(opP->con1))) {
 						tmpreg=0x38; /* 7.0 */
@@ -1598,7 +1662,7 @@ char	*instring;
 					if(isvar(opP->con1) &&
 					   !subs(opP->con1) &&
 					   seg(opP->con1)==SEG_TEXT &&
- 					   now_seg==SEG_TEXT &&
+					   now_seg==SEG_TEXT &&
 					   flagseen['m']==0 &&
 					    !index("~%&$?", s[0])) {
 						tmpreg=0x3A; /* 7.2 */
@@ -1725,28 +1789,29 @@ char	*instring;
 					    (the_ins.opcode[0] >= 0x6200) &&
 					    (the_ins.opcode[0] <= 0x6f00)) {
 					  add_frag(adds(opP->con1),offs(opP->con1),TAB(BCC68000,SZ_UNDEF));
-				        } else {
-					add_frag(adds(opP->con1),offs(opP->con1),TAB(BRANCH,SZ_UNDEF));
-				        }
+					} else {
+						add_frag(adds(opP->con1),offs(opP->con1),TAB(BRANCH,SZ_UNDEF));
+					}
 				} else {
 					/* JF:  This is the WRONG thing to do
 					add_frag((symbolS *)0,offs(opP->con1),TAB(BRANCH,BYTE)); */
 					the_ins.opcode[the_ins.numo-1]|=0xff;
+					offs(opP->con1)+=4;
 					add_fix('l',opP->con1,1);
 					addword(0);
-					addword(4);
+					addword(0);
 				}
 				break;
 			case 'w':
 				if(isvar(opP->con1)) {
 				    /* check for DBcc instruction */
-                                    if ((the_ins.opcode[0] & 0xf0f8) ==0x50c8){
-                                            /* size varies if patch */
-                                            /* needed for long form */
-                                            add_frag(adds(opP->con1),offs(opP->con1),TAB(DBCC,SZ_UNDEF));
-                                            break;
-                                        }
-                                        
+					if ((the_ins.opcode[0] & 0xf0f8) ==0x50c8) {
+					    /* size varies if patch */
+					    /* needed for long form */
+						add_frag(adds(opP->con1),offs(opP->con1),TAB(DBCC,SZ_UNDEF));
+						break;
+					}
+
 						/* Don't ask! */
 					opP->con1->e_exp.X_add_number+=2;
 					add_fix('w',opP->con1,1);
@@ -2127,7 +2192,7 @@ int in;
 	}
 	return out;
 }
-	
+
 void
 install_operand(mode,val)
 int mode;
@@ -2172,7 +2237,7 @@ int val;
 		the_ins.opcode[1]|=val<<5;
 		break;
 #endif
-		
+
 	case 't':
 		the_ins.opcode[1]|=(val<<10)|(val<<7);
 		break;
@@ -2270,7 +2335,7 @@ register struct m68k_op *opP;
 	if(c=='}')
 		c= *++str;		/* JF bitfield hack */
 	if(c) {
- 		c= *++str;
+		c= *++str;
 		if(!c)
 			as_bad("Missing operand");
 	}
@@ -2339,6 +2404,9 @@ char *str;
 			case 'b':
 				n=1;
 				break;
+			case '3':
+				n=2;
+				break;
 			case 'w':
 				n=2;
 				break;
@@ -2346,7 +2414,7 @@ char *str;
 				n=4;
 				break;
 			default:
-				as_fatal("confusing width %c",the_ins.reloc[m].wid);
+				as_fatal("Don't know how to figure width of %c in md_assemble()",the_ins.reloc[m].wid);
 			}
 
 			fix_new(frag_now,
@@ -2395,7 +2463,7 @@ char *str;
 			    the_ins.reloc[m].pcrel);
 		}
 		know(the_ins.fragb[n].fadd);
-                (void)frag_var(rs_machine_dependent,10,0,(relax_substateT)(the_ins.fragb[n].fragty),
+		(void)frag_var(rs_machine_dependent,10,0,(relax_substateT)(the_ins.fragb[n].fragty),
  the_ins.fragb[n].fadd,the_ins.fragb[n].foff,to_beg_P);
 	}
 	n=(the_ins.numo-the_ins.fragb[n-1].fragoff);
@@ -2472,7 +2540,7 @@ obstack_alloc(&robyn,sizeof(struct m68_incant));
 				slak->m_next=0;
 			slak=slak->m_next;
 		} while(slak);
-		
+
 		retval = hash_insert (op_hash, ins->name,(char *)hack);
 			/* Didn't his mommy tell him about null pointers? */
 		if(retval && *retval)
@@ -2496,6 +2564,9 @@ obstack_alloc(&robyn,sizeof(struct m68_incant));
 	alt_notend_table['#'] = 1;
 	alt_notend_table['f'] = 1;
 	alt_notend_table['F'] = 1;
+#ifdef REGISTER_PREFIX
+	alt_notend_table[REGISTER_PREFIX] = 1;
+#endif
 }
 
 #if 0
@@ -2512,6 +2583,7 @@ md_end()
 {
 }
 
+/* Equal to MAX_PRECISION in atof-ieee.c */
 #define MAX_LITTLENUMS 6
 
 /* Turn a string in input_line_pointer into a floating point constant of type
@@ -2528,7 +2600,7 @@ int *sizeP;
 	LITTLENUM_TYPE words[MAX_LITTLENUMS];
 	LITTLENUM_TYPE *wordP;
 	char	*t;
-	char	*atof_m68k();
+	char	*atof_ieee();
 
 	switch(type) {
 	case 'f':
@@ -2559,7 +2631,7 @@ int *sizeP;
 		*sizeP=0;
 		return "Bad call to MD_ATOF()";
 	}
-	t=atof_m68k(input_line_pointer,type,words);
+	t=atof_ieee(input_line_pointer,type,words);
 	if(t)
 		input_line_pointer=t;
 
@@ -2712,31 +2784,33 @@ register fragS *fragP;
     }
     break;
   case TAB(BCC68000,LONG):
-    {
 	/* only Bcc 68000 instructions can come here */
         /* change bcc into b!cc/jmp absl long */
 	fragP->fr_opcode[0] ^= 0x01; /* invert bcc */
         fragP->fr_opcode[1] = 0x6;   /* branch offset = 6 */
-        fragP->fr_opcode[2] = 0x4e;  /* put in jmp long (0x4ef9) */ 
-        fragP->fr_opcode[3] = 0xf9;  
+
+	/* JF: these used to be fr_opcode[2,3], but they may be in a
+	   different frag, in which case refering to them is a no-no.
+	   Only fr_opcode[0,1] are guaranteed to work. */
+        *buffer_address++ = 0x4e;  /* put in jmp long (0x4ef9) */ 
+        *buffer_address++ = 0xf9;  
         fragP->fr_fix += 2;	     /* account for jmp instruction */
         subseg_change(SEG_TEXT,0);
         fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, 
 					 fragP->fr_offset,0);
         fragP->fr_fix += 4;
         ext=0;
-    }
-    break;
+	break;
   case TAB(DBCC,LONG):
-    {
         /* only DBcc 68000 instructions can come here */
         /* change dbcc into dbcc/jmp absl long */
-        fragP->fr_opcode[2] = 0x00;  /* branch offset = 4 */
-        fragP->fr_opcode[3] = 0x04;  
-        fragP->fr_opcode[4] = 0x60;  /* put in bra pc+6 */ 
-        fragP->fr_opcode[5] = 0x06;  
-        fragP->fr_opcode[6] = 0x4e;  /* put in jmp long (0x4ef9) */ 
-        fragP->fr_opcode[7] = 0xf9;  
+	/* JF: these used to be fr_opcode[2-7], but that's wrong */
+        *buffer_address++ = 0x00;  /* branch offset = 4 */
+        *buffer_address++ = 0x04;  
+        *buffer_address++ = 0x60;  /* put in bra pc+6 */ 
+        *buffer_address++ = 0x06;  
+        *buffer_address++ = 0x4e;  /* put in jmp long (0x4ef9) */ 
+        *buffer_address++ = 0xf9;  
 
         fragP->fr_fix += 6;	     /* account for bra/jmp instructions */
         subseg_change(SEG_TEXT,0);
@@ -2744,7 +2818,6 @@ register fragS *fragP;
 					 fragP->fr_offset,0);
         fragP->fr_fix += 4;
         ext=0;
-    }
     break;
   case TAB(FBRANCH,SHORT):
     know((fragP->fr_opcode[1]&0x40)==0);
@@ -2764,7 +2837,7 @@ register fragS *fragP;
     subseg_change(SEG_TEXT,0);
     fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, fragP->fr_offset, 0);
     if((fragP->fr_opcode[1] & 0x3F) != 0x3A)
-    	as_bad("Internal error (long PC-relative operand) for insn 0x%04lx at 0x%lx",
+  	as_bad("Internal error (long PC-relative operand) for insn 0x%04lx at 0x%lx",
 	        fragP->fr_opcode[0],fragP->fr_address);
     fragP->fr_opcode[1]&= ~0x3F;
     fragP->fr_opcode[1]|=0x39;	/* Mode 7.1 */
@@ -2773,6 +2846,24 @@ register fragS *fragP;
 		       (long)(fragP->fr_symbol->sy_value + fragP->fr_offset),
 		       4); */
     ext=0;
+    break;
+  case TAB(PCLEA,SHORT):
+    subseg_change(SEG_TEXT,0);
+    fix_new(fragP,(int)(fragP->fr_fix),2,fragP->fr_symbol,(symbolS *)0,fragP->fr_offset,1);
+    fragP->fr_opcode[1] &= ~0x3F;
+    fragP->fr_opcode[1] |= 0x3A;
+    ext=2;
+    break;
+  case TAB(PCLEA,LONG):
+    subseg_change(SEG_TEXT,0);
+    fix_new(fragP,(int)(fragP->fr_fix)+2,4,fragP->fr_symbol,(symbolS *)0,fragP->fr_offset+2,1);
+    *buffer_address++ = 0x01;
+    *buffer_address++ = 0x70;
+    fragP->fr_fix+=2;
+    /* buffer_address+=2; */
+    ext=4;
+    break;
+
   }
   if(ext) {
     md_number_to_chars(buffer_address,(long)disp,(int)ext);
@@ -2789,102 +2880,72 @@ register fragS *fragP;
 int segtype;
 {
 	int	old_fix;
+	register char *buffer_address = fragP -> fr_fix + fragP -> fr_literal;
 
 	old_fix=fragP->fr_fix;
 
-        /* handle SZ_UNDEF first, it can be changed to BYTE or SHORT */
+	/* handle SZ_UNDEF first, it can be changed to BYTE or SHORT */
 	switch(fragP->fr_subtype) {
-        case TAB(DBCC,SZ_UNDEF):
-                if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
-                        fragP->fr_subtype=TAB(DBCC,SHORT);
-                        fragP->fr_var+=2;
-                        break;
-                }
-        	/* only DBcc 68000 instructions can come here */
-                /* change dbcc into dbcc/jmp absl long */
-                fragP->fr_opcode[2] = 0x00;  /* branch offset = 4 */
-                fragP->fr_opcode[3] = 0x04;  
-                fragP->fr_opcode[4] = 0x60;  /* put in bra pc+6 */ 
-                fragP->fr_opcode[5] = 0x06;  
-                fragP->fr_opcode[6] = 0x4e;  /* put in jmp long (0x4ef9) */ 
-                fragP->fr_opcode[7] = 0xf9;  
-                fragP->fr_fix += 6;	  /* account for bra/jmp instruction */
-                subseg_change(SEG_TEXT,0);
-                fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, 
-        					 fragP->fr_offset,0);
-                fragP->fr_fix += 4;
-                frag_wane(fragP);
-                break;
-                
-        case TAB(BCC68000,SZ_UNDEF):
-                if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
-                        fragP->fr_subtype=TAB(BCC68000,BYTE);
-                        break;
-                }
-        	/* only Bcc 68000 instructions can come here */
-                /* change bcc into b!cc/jmp absl long */
-        	fragP->fr_opcode[0] ^= 0x01; /* invert bcc */
-                fragP->fr_opcode[1] = 0x6;   /* branch offset = 6 */
-                fragP->fr_opcode[2] = 0x4e;  /* put in jmp long (0x4ef9) */ 
-                fragP->fr_opcode[3] = 0xf9;  
-                fragP->fr_fix += 2;	     /* account for jmp instruction */
-                subseg_change(SEG_TEXT,0);
-                fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, 
-        					 fragP->fr_offset,0);
-                fragP->fr_fix += 4;
-                frag_wane(fragP);
-                break;
-                
 	case TAB(BRANCH,SZ_UNDEF):
 		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
-                        fragP->fr_subtype=TAB(TABTYPE(fragP->fr_subtype),BYTE);
-                        break;
+			fragP->fr_subtype=TAB(TABTYPE(fragP->fr_subtype),BYTE);
+			break;
 		} else if(flagseen['m']) {
-		  if(fragP->fr_opcode[0]==0x61) {
-		    fragP->fr_opcode[0]= 0x4E;
-		    fragP->fr_opcode[1]= 0xB9;	/* JBSR with ABSL LONG offset */
-		    subseg_change(SEG_TEXT, 0);
-                    fix_new(fragP, fragP->fr_fix, 4, 
-			    fragP->fr_symbol, 0, fragP->fr_offset, 0);
-		    fragP->fr_fix+=4;
-		    frag_wane(fragP);
-                  } else if(fragP->fr_opcode[0]==0x60) {
-                    fragP->fr_opcode[0]= 0x4E;
-                    fragP->fr_opcode[1]= 0xF9;  /* JMP  with ABSL LONG offset */
-                    subseg_change(SEG_TEXT, 0);
-                    fix_new(fragP, fragP->fr_fix, 4, 
-			    fragP->fr_symbol, 0, fragP->fr_offset, 0);
-                    fragP->fr_fix+=4;
-                    frag_wane(fragP);
-		  } else {
-                    as_warn("Long branch offset to extern symbol not supported.");
-		  }
-		} else {	/* Symbol is still undefined.  Make it simple */
+			if(fragP->fr_opcode[0]==0x61) {
+				if(flagseen['l']) {
+					fragP->fr_opcode[0]= 0x4E;
+					fragP->fr_opcode[1]= 0xB8;	/* JBSR with ABSL WORD offset */
+					subseg_change(SEG_TEXT, 0);
+					fix_new(fragP, fragP->fr_fix, 2, 
+						fragP->fr_symbol, 0, fragP->fr_offset, 0);
+					fragP->fr_fix+=2;
+				} else {
+					fragP->fr_opcode[0]= 0x4E;
+					fragP->fr_opcode[1]= 0xB9;	/* JBSR with ABSL LONG offset */
+					subseg_change(SEG_TEXT, 0);
+					fix_new(fragP, fragP->fr_fix, 4, 
+						fragP->fr_symbol, 0, fragP->fr_offset, 0);
+					fragP->fr_fix+=4;
+				}
+				frag_wane(fragP);
+			} else if(fragP->fr_opcode[0]==0x60) {
+				if(flagseen['l']) {
+					fragP->fr_opcode[0]= 0x4E;
+					fragP->fr_opcode[1]= 0xF8;	/* JMP	with ABSL WORD offset */
+					subseg_change(SEG_TEXT, 0);
+					fix_new(fragP, fragP->fr_fix, 2, 
+						fragP->fr_symbol, 0, fragP->fr_offset, 0);
+					fragP->fr_fix+=2;
+				} else {
+					fragP->fr_opcode[0]= 0x4E;
+					fragP->fr_opcode[1]= 0xF9;	/* JMP	with ABSL LONG offset */
+					subseg_change(SEG_TEXT, 0);
+					fix_new(fragP, fragP->fr_fix, 4, 
+						fragP->fr_symbol, 0, fragP->fr_offset, 0);
+					fragP->fr_fix+=4;
+				}
+				frag_wane(fragP);
+			} else {
+				as_warn("Long branch offset to extern symbol not supported.");
+			}
+		} else if(flagseen['l']) {	/* Symbol is still undefined.  Make it simple */
+			fix_new(fragP,(int)(fragP->fr_fix),2,fragP->fr_symbol,
+ (symbolS *)0,fragP->fr_offset + 2,1);
+			fragP->fr_fix+=2;
+			fragP->fr_opcode[1]=0x00;
+			frag_wane(fragP);
+		} else {
 			fix_new(fragP,(int)(fragP->fr_fix),4,fragP->fr_symbol,
- (symbolS *)0,fragP->fr_offset+4,1);
+ (symbolS *)0,fragP->fr_offset + 4,1);
 			fragP->fr_fix+=4;
 			fragP->fr_opcode[1]=0xff;
 			frag_wane(fragP);
 			break;
 		}
-	default:
-                break;
-        }
-
-        /* now that SZ_UNDEF are taken care of, check others */
-        switch(fragP->fr_subtype) {
-        case TAB(BCC68000,BYTE):
-	case TAB(BRANCH,BYTE):
-			/* We can't do a short jump to the next instruction,
-			   so we force word mode.  */
-		if(fragP->fr_symbol && fragP->fr_symbol->sy_value==0 &&
- fragP->fr_symbol->sy_frag==fragP->fr_next) {
-                        fragP->fr_subtype=TAB(TABTYPE(fragP->fr_subtype),SHORT);
-			fragP->fr_var+=2;
-		}
 		break;
+
 	case TAB(FBRANCH,SZ_UNDEF):
-		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
+		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype || flagseen['l']) {
 			fragP->fr_subtype=TAB(FBRANCH,SHORT);
 			fragP->fr_var+=2;
 		} else {
@@ -2892,8 +2953,9 @@ int segtype;
 			fragP->fr_var+=4;
 		}
 		break;
+
 	case TAB(PCREL,SZ_UNDEF):
-		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
+		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype || flagseen['l']) {
 			fragP->fr_subtype=TAB(PCREL,SHORT);
 			fragP->fr_var+=2;
 		} else {
@@ -2901,6 +2963,101 @@ int segtype;
 			fragP->fr_var+=4;
 		}
 		break;
+
+	case TAB(BCC68000,SZ_UNDEF):
+		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
+			fragP->fr_subtype=TAB(BCC68000,BYTE);
+			break;
+		}
+		/* only Bcc 68000 instructions can come here */
+		/* change bcc into b!cc/jmp absl long */
+		fragP->fr_opcode[0] ^= 0x01; /* invert bcc */
+		if(flagseen['l']) {
+			fragP->fr_opcode[1] = 0x04;   /* branch offset = 6 */
+			/* JF: these were fr_opcode[2,3] */
+			buffer_address[0] = 0x4e;  /* put in jmp long (0x4ef9) */ 
+			buffer_address[1] = 0xf8;
+			fragP->fr_fix += 2;	     /* account for jmp instruction */
+			subseg_change(SEG_TEXT,0);
+			fix_new(fragP, fragP->fr_fix, 2, fragP->fr_symbol, 0, 
+							 fragP->fr_offset,0);
+			fragP->fr_fix += 2;
+		} else {
+			fragP->fr_opcode[1] = 0x06;   /* branch offset = 6 */
+			/* JF: these were fr_opcode[2,3] */
+			buffer_address[2] = 0x4e;  /* put in jmp long (0x4ef9) */ 
+			buffer_address[3] = 0xf9;
+			fragP->fr_fix += 2;	     /* account for jmp instruction */
+			subseg_change(SEG_TEXT,0);
+			fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, 
+							 fragP->fr_offset,0);
+			fragP->fr_fix += 4;
+		}
+		frag_wane(fragP);
+		break;
+
+	case TAB(DBCC,SZ_UNDEF):
+		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype) {
+			fragP->fr_subtype=TAB(DBCC,SHORT);
+			fragP->fr_var+=2;
+			break;
+		}
+		/* only DBcc 68000 instructions can come here */
+		/* change dbcc into dbcc/jmp absl long */
+		/* JF: these used to be fr_opcode[2-4], which is wrong. */
+		buffer_address[0] = 0x00;  /* branch offset = 4 */
+		buffer_address[1] = 0x04;  
+		buffer_address[2] = 0x60;  /* put in bra pc + ... */ 
+		if(flagseen['l']) {
+			/* JF: these were fr_opcode[5-7] */
+			buffer_address[3] = 0x04; /* plus 4 */
+			buffer_address[4] = 0x4e;/* Put in Jump Word */
+			buffer_address[5] = 0xf8;
+			fragP->fr_fix += 6;	  /* account for bra/jmp instruction */
+			subseg_change(SEG_TEXT,0);
+			fix_new(fragP, fragP->fr_fix, 2, fragP->fr_symbol, 0, 
+							 fragP->fr_offset,0);
+			fragP->fr_fix+=2;
+		} else {
+			/* JF: these were fr_opcode[5-7] */
+			buffer_address[3] = 0x06;  /* Plus 6 */
+			buffer_address[4] = 0x4e;  /* put in jmp long (0x4ef9) */ 
+			buffer_address[5] = 0xf9;  
+			fragP->fr_fix += 6;	  /* account for bra/jmp instruction */
+			subseg_change(SEG_TEXT,0);
+			fix_new(fragP, fragP->fr_fix, 4, fragP->fr_symbol, 0, 
+							 fragP->fr_offset,0);
+			fragP->fr_fix += 4;
+		}
+		frag_wane(fragP);
+		break;
+
+	case TAB(PCLEA,SZ_UNDEF):
+		if((fragP->fr_symbol->sy_type&N_TYPE)==segtype || flagseen['l']) {
+			fragP->fr_subtype=TAB(PCLEA,SHORT);
+			fragP->fr_var+=2;
+		} else {
+			fragP->fr_subtype=TAB(PCLEA,LONG);
+			fragP->fr_var+=6;
+		}
+		break;
+
+	default:
+		break;
+	}
+
+	/* now that SZ_UNDEF are taken care of, check others */
+	switch(fragP->fr_subtype) {
+	case TAB(BCC68000,BYTE):
+	case TAB(BRANCH,BYTE):
+			/* We can't do a short jump to the next instruction,
+			   so we force word mode.  */
+		if(fragP->fr_symbol && fragP->fr_symbol->sy_value==0 &&
+ fragP->fr_symbol->sy_frag==fragP->fr_next) {
+			fragP->fr_subtype=TAB(TABTYPE(fragP->fr_subtype),SHORT);
+			fragP->fr_var+=2;
+		}
+		break;
 	default:
 		break;
 	}
@@ -2920,7 +3077,7 @@ md_ri_to_chars(ri_p, ri)
      struct relocation_info *ri_p, ri;
 {
   unsigned char the_bytes[8];
-  
+
   /* this is easy */
   md_number_to_chars(the_bytes, ri.r_address, sizeof(ri.r_address));
   /* now the fun stuff */
@@ -2934,8 +3091,8 @@ md_ri_to_chars(ri_p, ri)
 }
 
 #ifndef WORKING_DOT_WORD
-int md_short_jump_size = 4;
-int md_long_jump_size = 6;
+const int md_short_jump_size = 4;
+const int md_long_jump_size = 6;
 
 void
 md_create_short_jump(ptr,from_addr,to_addr,frag,to_symbol)
@@ -3026,16 +3183,20 @@ int ok;
 	}
 
 	exp->e_siz=0;
-	if(/* ok!=80 && */exp->e_end[-1]==':' && (exp->e_end-exp->e_beg)>2) {
+	if(/* ok!=80 && */exp->e_end[-1]==':' && (exp->e_end-exp->e_beg)>=2) {
 		switch(exp->e_end[0]) {
 		case 's':
+		case 'S':
 		case 'b':
+		case 'B':
 			exp->e_siz=1;
 			break;
 		case 'w':
+		case 'W':
 			exp->e_siz=2;
 			break;
 		case 'l':
+		case 'L':
 			exp->e_siz=3;
 			break;
 		default:
@@ -3048,6 +3209,14 @@ int ok;
 	save_in=input_line_pointer;
 	input_line_pointer=exp->e_beg;
 	switch(expression(&(exp->e_exp))) {
+	case SEG_PASS1:
+		seg(exp)=SEG_ABSOLUTE;
+		adds(exp)=0;
+		subs(exp)=0;
+		offs(exp)= (ok==10) ? 1 : 0;
+		as_warn("Unknown expression: '%s' defaulting to %d",exp->e_beg,offs(exp));
+		break;
+
 	case SEG_NONE:
 		/* Do the same thing the VAX asm does */
 		seg(exp)=SEG_ABSOLUTE;
@@ -3205,7 +3374,7 @@ char ***vecP;
 		break;
 
 	case 'm':
-		/* Gas simply ignores this option! */
+		/* Gas almost ignores this option! */
 		(*argP)++;
 		if(**argP=='c')
 			(*argP)++;
@@ -3296,12 +3465,6 @@ char *str;
 	return 0;
 }
 
-print_address(add,fp)
-FILE *fp;
-{
-	fprintf(fp,"%#X",add);
-}
-
 #endif
 
 /* Possible states for relaxation:
